﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using Celeste;
using Celeste.Mod;
using FMOD.Studio;
using Microsoft.Xna.Framework;
using Monocle;
using MonoMod.Utils;
using YamlDotNet.Serialization;

namespace OneHourChallenge
{
    public class HCSaveData : EverestModuleSaveData
    {
        public bool HourChallengeModeEnabled { get; set; } = false;
        public int TimeRemaining { get; set; } = 216000;
        public int StartTime { get; set; } = 216000;
    }

    public class ChallengeModuleSettings : EverestModuleSettings
    {

        [SettingIgnore]
        public int BestScore { get; set; } = 0;

        [SettingIgnore]
        public int StartTime { get; set; } = 216000;
    }

    public class ChallengeModule : EverestModule
    {

        // dictionary that stores the state of each file
        public Dictionary<int, bool> HCFiles = new Dictionary<int, bool>();

        // Only one alive module instance can exist at any given time.
        public static ChallengeModule Instance;

        public ChallengeModule()
        {
            Instance = this;
        }
        // save data needed
        public override Type SaveDataType => typeof(HCSaveData);
        public static HCSaveData HMModuleSaveData => (HCSaveData)Instance._SaveData;

        public override Type SettingsType => typeof(ChallengeModuleSettings);

        public static ChallengeModuleSettings Settings => (ChallengeModuleSettings)Instance._Settings;

        // current file select slot (for OnHardcoreToggleSelected)
        private OuiFileSelectSlot currentSlot;

        // button to toggle hardcore mode (object because reflection is required)
        private object toggleButton;

        public override void LoadContent(bool firstLoad)
        {
        }

        // Optional, initialize anything after Celeste has initialized itself properly.
        public override void Initialize()
        {
        }


        public override void Load()
        {
            On.Celeste.SpeedrunTimerDisplay.DrawTime += SpeedrunTimerDisplay_DrawTime;
            On.Celeste.Level.Update += Level_Update;
            On.Celeste.OuiFileSelectSlot.Setup += On_OuiFileSelectSlot_Setup;
            On.Celeste.OuiFileSelectSlot.Render += On_OuiFileSelectSlot_Render;
            On.Celeste.OuiFileSelectSlot.Select += On_OuiFileSelectSlot_Select;
            On.Celeste.OuiFileSelectSlot.CreateButtons += On_OuiFileSelectSlot_CreateButtons;
            On.Celeste.OuiFileSelectSlot.OnNewGameSelected += On_OuiFileSelectSlot_OnNewGameSelected;
            On.Celeste.OuiJournal.Render += On_OuiJournal_Render;
            //On.Celeste.AreaComplete.Info += On_AreaComplete_Info;
            On.Celeste.SaveData.TryDelete += On_SaveData_TryDelete;
        }

        public Dictionary<int, int> IndexToTime = new Dictionary<int, int>()
        {
            { 0, 54000 }, // 15 minutes
            { 1, 108000 }, // 30 minutes
            { 2, 162000 }, // 45 minutes
            { 3, 216000 }, // 1 hour
            { 4, 432000 } // 2 hours
        };

        public Dictionary<int, int> TimeToIndex = new Dictionary<int, int>()
        {
            { 54000, 0 }, // 15 minutes
            { 108000, 1 }, // 30 minutes
            { 162000, 2 }, // 45 minutes
            { 216000, 3 }, // 1 hour
            { 432000, 4 } // 2 hours
        };

        public override void CreateModMenuSection(TextMenu menu, bool inGame, EventInstance snapshot)
        {
            base.CreateModMenuSection(menu, inGame, snapshot);
            menu.Add(new TextMenu.Slider(Dialog.Clean("ONEHOURCHALLENGE_TIME"), delegate (int i)
            {
                return IntToTime(IndexToTime[i]);
            }, 0, IndexToTime.Count - 1, TimeToIndex[Settings.StartTime]
            ).Change(v =>
            {
                Settings.StartTime = IndexToTime[v];
            }));
        }

        public static int CalcScore(SaveData data)
        {
            int score = 0;
            foreach (AreaStats areaStats in data.Areas_Safe)
            {
                for (int i = 0; i < areaStats.Modes.Length; i++)
                {
                    if (areaStats.Modes[i] != null)
                    {
                        // Hearts
                        if (areaStats.Modes[i].HeartGem)
                        {
                            score += 10 * (i + 1);
                        }
                        // Completed Levels
                        score += areaStats.Modes[i].SingleRunCompleted ? 30 * (i+1) : 0;
                    }
                }
                // Berries
                score += areaStats.TotalStrawberries * 5;
                // Cassettes
                score += areaStats.Cassette ? 10 : 0;
            }
            return score;
        }

        private void SpeedrunTimerDisplay_DrawTime(On.Celeste.SpeedrunTimerDisplay.orig_DrawTime orig, Vector2 position, string timeString, float scale, bool valid, bool finished, bool bestTime, float alpha)
        {
            if (HMModuleSaveData.HourChallengeModeEnabled)
            {
                if (timeString.StartsWith("Score: "))
                {
                    orig(position, timeString, scale, valid, finished, bestTime, alpha);
                    return;
                }
                if (timeString == "Time's Up!")
                {
                    orig(position, timeString, scale, valid, finished, bestTime, alpha);
                    return;
                }
                else
                {
                    if (numberWidth == -1f)
                    {
                        getNumberWidth();
                    }
                    int time = HMModuleSaveData.TimeRemaining;
                    timeString = IntToTime(time);
                    int score = CalcScore(SaveData.Instance);
                    string str = $"Score: {score}";
                    Vector2 scorePos = HMModuleSaveData.TimeRemaining > 0 ? position + Vector2.UnitY * 50f : new Vector2(1920f / 2 - (numberWidth * str.Length / 2), 1080f / 2);
                    SpeedrunTimerDisplay.DrawTime(scorePos, str, scale, true, false, (score > Settings.BestScore ? true : false), alpha);
                    //if (HMModuleSaveData.TimeRemaining < 1)
                    //SpeedrunTimerDisplay.DrawTime(scorePos - Vector2.UnitY * 50f, "Time's Up!", scale, true, false, (score > Settings.BestScore ? true : false), alpha);
                }
                
            }
            orig(position, timeString, scale, valid, finished, bestTime, alpha);
        }
        private static void getNumberWidth()
        {
            numberWidth = (float)typeof(SpeedrunTimerDisplay).GetField("numberWidth", BindingFlags.NonPublic | BindingFlags.Static).GetValue(null);
        }
        private static float numberWidth = -1f;

        private void Level_Update(On.Celeste.Level.orig_Update orig, Level self)
        {
            orig(self);
            if (!self.Completed && HMModuleSaveData.HourChallengeModeEnabled)
            {
                if (HMModuleSaveData.TimeRemaining < 1)
                {
                    // TIME'S UP
                    foreach (Player player in self.Tracker.GetEntities<Player>())
                    {
                        // Freeze players
                        player.Speed = Vector2.Zero;
                        player.StateMachine.State = Player.StDummy;
                    }
                    {
                        Player player;
                        if ((player = self.Tracker.GetEntity<Player>()) == null)
                        {
                            return;
                        }
                        
                        PlayerDeadBody body = player.Die(Vector2.Zero, true, false);
                        body.DeathAction = () =>
                        {
                            int CurScore = CalcScore(SaveData.Instance);
                            bool newBest = false;
                            if (Settings.BestScore < CurScore)
                            {
                                Settings.BestScore = CurScore;
                                newBest = true;
                            }
                            Engine.TimeRate = 1f;
                            self.Session.InArea = false;
                            Audio.BusStopAll("bus:/gameplay_sfx", immediate: true);
                            Engine.Scene = new GameOverScreen(newBest);
                        };
                    }
                } else
                {
                    HMModuleSaveData.TimeRemaining -= 1;
                }
            }
        }

        public override void Unload()
        {
            On.Celeste.OuiFileSelectSlot.Setup -= On_OuiFileSelectSlot_Setup;
            On.Celeste.OuiFileSelectSlot.Render -= On_OuiFileSelectSlot_Render;
            On.Celeste.OuiFileSelectSlot.Select -= On_OuiFileSelectSlot_Select;
            On.Celeste.OuiFileSelectSlot.CreateButtons -= On_OuiFileSelectSlot_CreateButtons;
            On.Celeste.OuiFileSelectSlot.OnNewGameSelected -= On_OuiFileSelectSlot_OnNewGameSelected;
            On.Celeste.OuiJournal.Render -= On_OuiJournal_Render;
            //On.Celeste.AreaComplete.Info -= On_AreaComplete_Info;
            On.Celeste.SaveData.TryDelete -= On_SaveData_TryDelete;
        }

        /// <summary>
        /// Reads a file slot's dictionary entry to determine whether it is in 1 Hour Challenge mode.
        /// If the entry is not present in the dictionary, it will add it.
        /// </summary>
        /// <param name="slot">The slot number.</param>
        /// <returns>True if the file is in 1 Hour Challenge mode, false otherwise.</returns>
        public bool Is1HCFile(int slot)
        {
            // attempt to get from dictionary
            if (HCFiles.TryGetValue(slot, out bool value))
            {
                return value;
            }
            // default to false if the save file doesn't exist at all
            else if (!UserIO.Exists(SaveData.GetFilename(slot)))
            {
                return false;
            }
            else
            {
                // attempt to load from save file
                try
                {
                    Logger.Log(LogLevel.Info, "1 Hour Challenge", $"Getting save data from file {slot}");
                    LoadSaveData(slot);
                    HCFiles[slot] = HMModuleSaveData.HourChallengeModeEnabled;
                    return HCFiles[slot];
                }
                // default to false if that fails somehow
                catch
                {
                    Logger.Log(LogLevel.Warn, "1 Hour Challenge", $"Could not get save data from file {slot}");
                    HCFiles[slot] = false;
                    return false;
                }
            }
        }

        /// <summary>
        /// Resets a file's Hour Challenge Mode status when it is successfully deleted.
        /// </summary>
        /// <param name="orig">The original method.</param>
        /// <param name="slot">The ID of the file being deleted.</param>
        /// <returns>The result of the original method (whether the file was successfully deleted).</returns>
        public bool On_SaveData_TryDelete(On.Celeste.SaveData.orig_TryDelete orig, int slot)
        {
            bool result = orig(slot);
            if (result)
            {
                HCFiles[slot] = false;
            }
            return result;
        }

        public void On_OuiFileSelectSlot_Setup(On.Celeste.OuiFileSelectSlot.orig_Setup orig, OuiFileSelectSlot self)
        {
            orig(self);
            if (Is1HCFile(self.FileSlot))
            {
                //self.Portrait.Play(hardcorePortrait);
            }
        }

        /// <summary>
        /// Draws the hardcore tab if that file is in hardcore mode.
        /// </summary>
        /// <param name="orig">The original method.</param>
        /// <param name="self">The file select slot we are drawing on.</param>
        public void On_OuiFileSelectSlot_Render(On.Celeste.OuiFileSelectSlot.orig_Render orig, OuiFileSelectSlot self)
        {
            if (Is1HCFile(self.FileSlot))
            {
                // mimic the original code for drawing tabs
                DynData<OuiFileSelectSlot> data = new DynData<OuiFileSelectSlot>(self);
                float highlightEase = (float)GetInstanceField(self, "highlightEase");
                float newgameFade = (float)GetInstanceField(self, "newgameFade");
                float scaleFactor = Ease.CubeInOut(highlightEase);
                bool selected = data.Get<bool>("highlighted");
                Vector2 pos = selected ? self.Position + Vector2.UnitX * 700f * scaleFactor : self.Position;
                Vector2 vector = pos - new Vector2(scaleFactor * 360f, 0f);// ((Vector2.UnitX * scaleFactor * 360f));
                float scale = self.Exists ? 1f : newgameFade;
                if (!self.Corrupted && (newgameFade > 0f || self.Exists))
                {
                    MTN.FileSelect["1hctab"].DrawCentered(vector, Color.White * scale, 1f, Calc.ToRad(180));
                }
            }
            orig(self);
        }
        PropertyInfo OuiFileSelectSlot_selected = typeof(OuiFileSelectSlot).GetProperty("selected", bindFlags);
        public void On_OuiFileSelectSlot_Select(On.Celeste.OuiFileSelectSlot.orig_Select orig, OuiFileSelectSlot self, bool resetButtonIndex)
        {
            if (!self.Exists && resetButtonIndex)
            {
                HCFiles[self.FileSlot] = false;
                self.Portrait.Play("idle_normal");
            }
            orig(self, resetButtonIndex);
        }

        /// <summary>
        /// Adds the "1HourChallenge" button to file creation.
        /// </summary>
        /// <param name="orig">The original method.</param>
        /// <param name="self">The file select slot we are creating.</param>
        public void On_OuiFileSelectSlot_CreateButtons(On.Celeste.OuiFileSelectSlot.orig_CreateButtons orig, OuiFileSelectSlot self)
        {
            orig(self);
            currentSlot = self;
            if (!self.Exists)
            {
                // mess of reflection necessary because the Button class is private
                Type buttonClass = GetNestedType(typeof(OuiFileSelectSlot), "Button");
                toggleButton = Activator.CreateInstance(buttonClass);
                string dialogId = HCFiles[self.FileSlot] ? "FILE_HOURCHALLENGE_ON" : "FILE_HOURCHALLENGE_OFF";
                SetInstanceField(toggleButton, "Label", Dialog.Clean(dialogId));
                SetInstanceField(toggleButton, "Action", (Action)OnHCToggleSelected);
                SetInstanceField(toggleButton, "Scale", 0.7f);
                Type t = typeof(List<>).MakeGenericType(buttonClass);
                var buttons = Convert.ChangeType(GetInstanceField(self, "buttons"), t);
                buttons.GetType().GetMethod("Add").Invoke(buttons, new object[] { toggleButton });
                SetInstanceField(self, "buttons", buttons);
            }
        }

        private void OnHCToggleSelected()
        {
            int index = currentSlot.FileSlot;
            if (!Is1HCFile(index))
            {
                HCFiles[index] = true;
                SetInstanceField(toggleButton, "Label", Dialog.Clean("FILE_HOURCHALLENGE_ON"));
                Audio.Play("event:/ui/main/button_toggle_on");
                //currentSlot.Portrait.Play(hardcorePortrait);
            }
            else
            {
                HCFiles[index] = false;
                SetInstanceField(toggleButton, "Label", Dialog.Clean("FILE_HOURCHALLENGE_OFF"));
                Audio.Play("event:/ui/main/button_toggle_off");
                //currentSlot.Portrait.Play("idle_normal");
            }
        }

        /// <summary>
        /// Writes the mod's SaveData to a newly created save file.
        /// </summary>
        /// <param name="orig">The original method.</param>
        /// <param name="self">The file select slot being created.</param>
        public void On_OuiFileSelectSlot_OnNewGameSelected(On.Celeste.OuiFileSelectSlot.orig_OnNewGameSelected orig, OuiFileSelectSlot self)
        {
            LoadSaveData(self.FileSlot);
            HMModuleSaveData.HourChallengeModeEnabled = HCFiles[self.FileSlot];
            HMModuleSaveData.TimeRemaining = Settings.StartTime;
            HMModuleSaveData.StartTime = Settings.StartTime;
            SaveSaveData(self.FileSlot);
            orig(self);
        }

        public static string IntToTime(int time)
        {
            string minutes = ((time / 3600) % 60).ToString();
            string seconds = ((time / 60) % 60).ToString();
            string hours = (time / 216000).ToString();
            if (minutes.Length == 1) minutes = "0" + minutes;
            if (seconds.Length == 1) seconds = "0" + seconds;
            if ((time / 216000) > 0)
            {
                return $"{hours}:{minutes}.{seconds}";
            }
            return $"{minutes}.{seconds}";
        }

        /// <summary>
        /// Shows the hardcore tab on the journal screen.
        /// </summary>
        /// <param name="orig">The original method.</param>
        /// <param name="self">The journal we are drawing on.</param>
        public void On_OuiJournal_Render(On.Celeste.OuiJournal.orig_Render orig, OuiJournal self)
        {
            Vector2 vector = self.Position + new Vector2(128f, 120f);
            if (HMModuleSaveData.HourChallengeModeEnabled)
            {
                MTN.FileSelect["1hctab"].DrawCentered(vector + new Vector2(80f, 425f), Color.White, 1f, (float)Math.PI / 2f);
            }
            orig(self);
        }

        // ================ Reflection helpers ================

        private static readonly BindingFlags bindFlags = BindingFlags.Instance
            | BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static;

        /// <summary>
        /// Gets the value of a field from a class instance, even if it is non-public.
        /// </summary>
        /// <param name="instance">The instance of the class to get the field value from.</param>
        /// <param name="fieldName">The name of the field to request.</param>
        /// <returns>The value of the field requested.</returns>
        public static object GetInstanceField(object instance, string fieldName)
        {
            FieldInfo field = instance.GetType().GetField(fieldName, bindFlags);
            return field.GetValue(instance);
        }

        /// <summary>
        /// Sets the value of a field from a class instance, even if it is non-public.
        /// </summary>
        /// <param name="instance">The instance of the class to set the field of.</param>
        /// <param name="fieldName">The name of the field to change.</param>
        /// <param name="value">The value to set the field to.</param>
        public static void SetInstanceField(object instance, string fieldName, object value)
        {
            FieldInfo field = instance.GetType().GetField(fieldName, bindFlags);
            field.SetValue(instance, value);
        }

        /// <summary>
        /// Calls a method from a class instance, even if it is non-public.
        /// </summary>
        /// <param name="instance">The instance of the class to call a method from.</param>
        /// <param name="methodName">The name of the method to be called.</param>
        /// <param name="parameters">An array of parameters to call the method with.</param>
        /// <returns>The result of the method (or null if the method is void).</returns>
        public static object CallInstanceMethod(object instance, string methodName, object[] parameters = null)
        {
            MethodInfo method = instance.GetType().GetMethod(methodName, bindFlags);
            return method.Invoke(instance, parameters);
        }

        /// <summary>
        /// Gets the type of a class that is nested in another class.
        /// </summary>
        /// <param name="baseClass">The class that contains the type requested.</param>
        /// <param name="name">The name of the type to request.</param>
        /// <returns>A Type object corresponding to the requested type.</returns>
        public static Type GetNestedType(Type baseClass, string name)
        {
            return baseClass.GetNestedType(name, bindFlags);
        }

    }
}
